/*
 * Copyright (C) 2022  Aleksei Bavshin
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef QGREETDPROTO_HPP_
#define QGREETDPROTO_HPP_

#include <optional>
#include <variant>

#include <QString>
#include <QtCore>

QT_BEGIN_NAMESPACE
class QJsonObject;
class QDataStream;
QT_END_NAMESPACE

namespace QGreetd {
Q_NAMESPACE

/*!
 * CreateSession initiates a login attempt for the given user.
 * CreateSession returns either a Response::AuthMessage,
 * Response::Success or Response::Failure.
 *
 * If an auth message is returned, it should be answered with a
 * Request::PostAuthMessageResponse. If a success is returned, the session
 * can then be started with Request::StartSession.
 *
 * If a login flow needs to be aborted at any point, send
 * Request::CancelSession. Note that the session is cancelled
 * automatically on error.
 */
struct CreateSession {
    QString username;
};

/*!
 * PostAuthMessageResponse responds to the last auth message, and returns
 * either a Response::AuthMessage, Response::Success or Response::Failure.
 *
 * If an auth message is returned, it should be answered with a
 * Request::PostAuthMessageResponse. If a success is returned, the session
 * can then be started with Request::StartSession.
 */
struct PostAuthMessageResponse {
    std::optional<QString> response;
};

/*!
 * Start a successfully logged in session. This will fail if the session
 * has pending messages or has encountered an error.
 */
struct StartSession {
    QStringList cmd;
    QStringList env;
};

/*!
 * Cancel a session. This can only be done if the session has not been
 * started. Cancel does not have to be called if an error has been
 * encountered in its setup or login flow.
 */
struct CancelSession {};

/*!
 * A request from a greeter to greetd. The request type is internally tagged
 * with the"type" field, with the type written in snake_case.
 *
 * Example serialization:
 *
 * ```json
 * {
 *    "type": "create_session",
 *    "username": "bob"
 * }
 * ```
 */
typedef std::variant<CreateSession, PostAuthMessageResponse, StartSession, CancelSession> Request;

/*!
 * The request was successful.
 */
struct Success {};

enum class ErrorType { Error, AuthError };

/*!
 * The request failed. See the type and/or description for more
 * information about this failure.
 */
struct Error {
    ErrorType error_type;
    QString description;
};

enum class AuthMessageType {
    Visible,
    Secret,
    Info,
    Error,
};

/*!
 * An authentication message needs to be answered to continue through the
 * authentication flow.
 *
 * An authentication message can consist of anything. While it will
 * commonly just be a request for the users' password, it could also ask
 * for TOTP codes, or whether or not you felt sad when Littlefoot's mother
 * died in the original "Land Before Time". It is therefore important that
 * no assumptions are made about the questions that will be asked, and
 * attempts to automatically answer these questions should not be made.
 */
struct AuthMessage {
    AuthMessageType auth_message_type;
    QString auth_message;
};

/*!
 * A response from greetd to a greeter. The request type is internally tagged
 * with the"type" field, with the type written in snake_case.
 *
 * Example serialization:
 *
 * ```json
 * {
 *    "type": "auth_message",
 *    "message": "Password:",
 *    "message_type": "secret"
 * }
 * ```
 */
typedef std::variant<Success, Error, AuthMessage> Response;

Q_ENUM_NS(ErrorType);
Q_ENUM_NS(AuthMessageType);

QDataStream &operator>>(QDataStream &, Request &);
QDataStream &operator>>(QDataStream &, Response &);

#define QGREETD_DECLARE_CONVERSIONS(Type)                                                          \
    void from_json(const QJsonObject &, Type &);                                                   \
    void to_json(QJsonObject &, const Type &);                                                     \
    QDataStream &operator<<(QDataStream &, const Type &);

QGREETD_DECLARE_CONVERSIONS(CreateSession);
QGREETD_DECLARE_CONVERSIONS(PostAuthMessageResponse);
QGREETD_DECLARE_CONVERSIONS(StartSession);
QGREETD_DECLARE_CONVERSIONS(CancelSession);
QGREETD_DECLARE_CONVERSIONS(Request);
QGREETD_DECLARE_CONVERSIONS(Success);
QGREETD_DECLARE_CONVERSIONS(Error);
QGREETD_DECLARE_CONVERSIONS(AuthMessage);
QGREETD_DECLARE_CONVERSIONS(Response);
#undef QGREETD_DECLARE_CONVERSIONS

} // namespace QGreetd

#endif /* QGREETDPROTO_HPP_ */
