/*
 * Copyright (C) 2022  Aleksei Bavshin
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "QGreetd/qgreetdclient.hpp"

#include <QDataStream>
#include <QDir>
#include <QJsonObject>
#include <QProcessEnvironment>
#include <QSysInfo>

namespace QGreetd {
Q_LOGGING_CATEGORY(greetdClient, "qgreetd.client")

QString QGreetdClient::defaultPath() {
    auto env = QProcessEnvironment::systemEnvironment();
    return env.value("GREETD_SOCK");
}

QGreetdClient::QGreetdClient(QObject *parent) : QObject(parent), sock{new QLocalSocket(this)} {
    /* greetd protocol uses native byte order */
    stream.setByteOrder(static_cast<QDataStream::ByteOrder>(QSysInfo::ByteOrder));
    stream.setDevice(sock);

    connect(sock, &QLocalSocket::connected, this, &QGreetdClient::connected);
    connect(sock, &QLocalSocket::readyRead, this, &QGreetdClient::readResponse);
    connect(sock, &QLocalSocket::disconnected, this, &QGreetdClient::disconnected);
    connect(sock, &QLocalSocket::errorOccurred, this, &QGreetdClient::handleError);
    connect(this, &QGreetdClient::response, this, &QGreetdClient::handleResponse);
}

void QGreetdClient::connectToDaemon(QString path) {
    if (path.isEmpty()) {
        emit disconnected();
    }

    QDir dir = QDir::currentPath();
    path = dir.absoluteFilePath(path);
    sock->connectToServer(path);
}

void QGreetdClient::readResponse() {
    Response resp;
    stream.startTransaction();
    stream >> resp;

    if (!stream.commitTransaction()) {
        qCDebug(greetdClient) << "Incomplete read buffer";
        return;
    }

    emit response(resp);
}

void QGreetdClient::handleError(QLocalSocket::LocalSocketError socketError) {
    qWarning() << "Socket disconnected with" << socketError;

    emit disconnected();
}

template <> void QGreetdClient::handleResponseImpl<Success>(const Success &) {
    switch (state) {
    case State::Authenticated:
        qCInfo(greetdClient) << "Session started";

        emit done();
        break;

    case State::Cancelled:
        qCDebug(greetdClient) << "Session cancelled";

        state = State::Initial;
        emit reset();
        break;

    default:
        qCInfo(greetdClient) << "Authentication succeeded," << state;

        state = State::Authenticated;
        emit authenticationComplete();
        break;
    }
}

template <> void QGreetdClient::handleResponseImpl<Error>(const Error &err) {
    switch (state) {
    case State::Cancelled:
        qCInfo(greetdClient) << "Ignoring CancelSession error:" << err.error_type << "-"
                             << err.description;

        state = State::Initial;
        emit reset();
        break;

    default:
        qCInfo(greetdClient) << "Session cancelled due to" << err.error_type << "-"
                             << err.description;

        emit showMessage(err.description, MessageType::Error);
        cancelAuthentication();
        break;
    }
}

template <> void QGreetdClient::handleResponseImpl<AuthMessage>(const AuthMessage &msg) {
    qCDebug(greetdClient) << "AuthMessage received: " << msg.auth_message_type << "-"
                          << msg.auth_message;

    switch (msg.auth_message_type) {
    case AuthMessageType::Visible:
        emit showPrompt(msg.auth_message, PromptType::Visible);
        break;
    case AuthMessageType::Secret:
        emit showPrompt(msg.auth_message, PromptType::Secret);
        break;
    case AuthMessageType::Info:
        emit showMessage(msg.auth_message, MessageType::Info);
        respond();
        break;
    case AuthMessageType::Error:
        emit showMessage(msg.auth_message, MessageType::Error);
        respond();
        break;
    }
}

void QGreetdClient::handleResponse(const Response &response) {
    std::visit([this](const auto &variant) { return handleResponseImpl(variant); }, response);
}

void QGreetdClient::request(const Request &req) {
    Q_ASSERT(sock->state() == QLocalSocket::LocalSocketState::ConnectedState);
    stream << req;
};

void QGreetdClient::authenticate(const QString &username) {
    qCDebug(greetdClient) << "Creating new session for user" << username;

    request(CreateSession{username});
}

void QGreetdClient::respond(std::optional<QString> response) {
    request(PostAuthMessageResponse{response});
}

void QGreetdClient::startSession(const QStringList &cmd, QStringList env) {
    qCDebug(greetdClient) << "Starting session" << cmd;

    request(StartSession{cmd, env});
}

void QGreetdClient::cancelAuthentication() {
    qCDebug(greetdClient) << "Cancelling session";

    state = State::Cancelled;
    request(CancelSession{});
}

} // namespace QGreetd
