/*
 * Copyright (C) 2022  Aleksei Bavshin
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "QGreetd/qgreetdprotocol.hpp"

#include <QDataStream>
#include <QJsonArray>
#include <QJsonDocument>
#include <QJsonObject>
#include <stdexcept>

namespace QGreetd {

template <typename T> inline T read_helper(const QJsonObject &json) {
    T val;
    from_json(json, val);
    return val;
}

#define GET_OPTIONAL_FIELD(type, name)                                                             \
    do {                                                                                           \
        if (auto value = json.value(#name); value.is##type()) {                                    \
            obj.name = value.to##type();                                                           \
        }                                                                                          \
    } while (0)

#define GET_REQUIRED_FIELD(type, name)                                                             \
    do {                                                                                           \
        if (auto value = json.value(#name); value.is##type()) {                                    \
            obj.name = value.to##type();                                                           \
        } else {                                                                                   \
            throw std::runtime_error("Required field \"" #name "\" is missing ");                  \
        }                                                                                          \
    } while (0)

void from_json(const QJsonObject &json, CreateSession &obj) {
    GET_REQUIRED_FIELD(String, username);
}

void to_json(QJsonObject &json, const CreateSession &obj) {
    json["type"] = "create_session";
    json["username"] = obj.username;
}

void from_json(const QJsonObject &json, PostAuthMessageResponse &obj) {
    GET_OPTIONAL_FIELD(String, response);
}

void to_json(QJsonObject &json, const PostAuthMessageResponse &obj) {
    json["type"] = "post_auth_message_response";
    if (obj.response.has_value()) {
        json["response"] = obj.response.value();
    }
}

void from_json(const QJsonObject &json, StartSession &obj) {
    if (auto value = json.value("cmd"); value.isArray()) {
        obj.cmd = value.toVariant().toStringList();
    }
    if (auto value = json.value("env"); value.isArray()) {
        obj.env = value.toVariant().toStringList();
    }
}

void to_json(QJsonObject &json, const StartSession &obj) {
    json["type"] = "start_session";
    json["cmd"] = QJsonArray::fromStringList(obj.cmd);
    json["env"] = QJsonArray::fromStringList(obj.env);
}
void from_json(const QJsonObject &json, CancelSession &obj) {}

void to_json(QJsonObject &json, const CancelSession &obj) {
    json["type"] = "cancel_session";
}

void from_json(const QJsonObject &json, Request &request) {
    auto type = json.value("type").toString().toLower();
    if (type == "create_session") {
        request = read_helper<CreateSession>(json);
    } else if (type == "post_auth_message_response") {
        request = read_helper<PostAuthMessageResponse>(json);
    } else if (type == "start_session") {
        request = read_helper<StartSession>(json);
    } else if (type == "cancel_session") {
        request = read_helper<CancelSession>(json);
    } else {
        throw std::runtime_error{"Unsupported Request type: " + type.toStdString()};
    }
}

void from_json(const QJsonObject &json, Success &obj){
    // pass
};

void to_json(QJsonObject &json, const Success &obj) {
    json["type"] = "success";
}

static QMap<QString, ErrorType> error_type_map = {
    {"error", ErrorType::Error},
    {"auth_error", ErrorType::AuthError},
};

void from_json(const QJsonObject &json, Error &obj) {
    auto type = json.value("error_type").toString().toLower();
    if (auto it = error_type_map.find(type); it != error_type_map.end()) {
        obj.error_type = it.value();
    } else {
        throw std::runtime_error{"Unsupported enum ErrorType value: " + type.toStdString()};
    }

    GET_REQUIRED_FIELD(String, description);
};

void to_json(QJsonObject &json, const Error &obj) {
    json["type"] = "error";
    json["error_type"] = error_type_map.key(obj.error_type);
    json["description"] = obj.description;
}

static QMap<QString, AuthMessageType> auth_message_type_map = {
    {"visible", AuthMessageType::Visible},
    {"secret", AuthMessageType::Secret},
    {"info", AuthMessageType::Info},
    {"error", AuthMessageType::Error},
};

void from_json(const QJsonObject &json, AuthMessage &obj) {
    auto type = json.value("auth_message_type").toString().toLower();
    if (auto it = auth_message_type_map.find(type); it != auth_message_type_map.end()) {
        obj.auth_message_type = it.value();
    } else {
        throw std::runtime_error{"Unsupported enum AuthMessageType value: " + type.toStdString()};
    }

    GET_REQUIRED_FIELD(String, auth_message);
}

void to_json(QJsonObject &json, const AuthMessage &obj) {
    json["type"] = "auth_message";
    json["auth_message_type"] = auth_message_type_map.key(obj.auth_message_type);
    json["auth_message"] = obj.auth_message;
}

void to_json(QJsonObject &json, const Request &request) {
    std::visit([&json](auto arg) { to_json(json, arg); }, request);
}

void from_json(const QJsonObject &json, Response &response) {
    auto type = json.value("type").toString().toLower();
    if (type == "success") {
        response = read_helper<Success>(json);
    } else if (type == "error") {
        response = read_helper<Error>(json);
    } else if (type == "auth_message") {
        response = read_helper<AuthMessage>(json);
    } else {
        throw std::runtime_error{"Unhandled Response type: " + type.toStdString()};
    }
}

void to_json(QJsonObject &json, const Response &response) {
    std::visit([&json](auto arg) { to_json(json, arg); }, response);
}

QJsonDocument readObject(QDataStream &in) {
    char *data;
    uint len;

    in.startTransaction();
    in.readBytes(data, len);

    QScopedArrayPointer<char> cleanup{data};

    if (in.status() != QDataStream::Ok) {
        in.rollbackTransaction();
        return {};
    }

    in.commitTransaction();
    return QJsonDocument::fromJson(QByteArray::fromRawData(data, len));
}

QDataStream &operator>>(QDataStream &in, Request &request) {
    if (auto doc = readObject(in); doc.isObject()) {
        from_json(doc.object(), request);
    }
    return in;
}

QDataStream &operator>>(QDataStream &in, Response &response) {
    if (auto doc = readObject(in); doc.isObject()) {
        from_json(doc.object(), response);
    }
    return in;
}

#define STREAM_OUTPUT_IMPL(Type)                                                                   \
    QDataStream &operator<<(QDataStream &out, const Type &obj) {                                   \
        QJsonObject json;                                                                          \
        to_json(json, obj);                                                                        \
        QByteArray bytes = QJsonDocument{json}.toJson(QJsonDocument::JsonFormat::Compact);         \
        return out.writeBytes(bytes.data(), bytes.size());                                         \
    }

STREAM_OUTPUT_IMPL(CreateSession);
STREAM_OUTPUT_IMPL(PostAuthMessageResponse);
STREAM_OUTPUT_IMPL(StartSession);
STREAM_OUTPUT_IMPL(CancelSession);
STREAM_OUTPUT_IMPL(Request);
STREAM_OUTPUT_IMPL(Success);
STREAM_OUTPUT_IMPL(Error);
STREAM_OUTPUT_IMPL(AuthMessage);
STREAM_OUTPUT_IMPL(Response);
#undef STREAM_OUT_IMPL

} // namespace QGreetd
