/*
 * Copyright (C) 2022  Aleksei Bavshin
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "qgreetercontext.hpp"
#include "config.h"
#include "util/configuration.hpp"
#include <QCommandLineParser>
#include <QCoreApplication>
#include <QHostInfo>

namespace {
QStringList findDataDirs(const QLatin1String &subdir) {
    return QStandardPaths::locateAll(QStandardPaths::GenericDataLocation, subdir,
                                     QStandardPaths::LocateDirectory);
}
} // namespace

QMLGREET_DECLARE_OPTION(General, RememberLastUser,
                        QCommandLineOption("remember-user", "Remember last logged-in username."),
                        false);
QMLGREET_DECLARE_OPTION(General, RememberLastSession,
                        QCommandLineOption("remember-session", "Remember last selected session."),
                        false);

QMLGREET_DECLARE_OPTION(
    Wayland, SessionDir,
    QCommandLineOption("wayland-sessions",
                       "Path to a wayland session directory (can be specified multiple times).",
                       "path"),
    findDataDirs(QLatin1String{"wayland-sessions"}));

QMLGREET_DECLARE_OPTION(
    X11, SessionDir,
    QCommandLineOption("xsessions",
                       "Path to an X11 session directory (can be specified multiple times).",
                       "path"),
    findDataDirs(QLatin1String{"xsessions"}));

namespace QmlGreet {

Q_LOGGING_CATEGORY(greeterContext, "qmlgreet.context");

QGreeterContext::QGreeterContext(QObject *parent)
    : QGreetd::QGreetdClient{parent},
      m_sessionModel{new SessionModel{this}},
      m_userSettings{QSettings::UserScope, "qmlgreet"},
      m_hostname{QHostInfo::localHostName()},
      m_currentSession{0} {

    for (auto dir : Settings::Wayland::SessionDir.getList()) {
        m_sessionModel->loadSessions(SessionModel::Wayland, dir);
    }

    for (auto dir : Settings::X11::SessionDir.getList()) {
        m_sessionModel->loadSessions(SessionModel::X11, dir);
    }

    connect(this, &QGreetd::QGreetdClient::authenticationComplete, this,
            &QGreeterContext::onAuthenticationComplete);

    if (Settings::General::RememberLastUser || Settings::General::RememberLastSession) {
        connect(this, &QGreetd::QGreetdClient::done, [this]() {
            qCDebug(greeterContext) << "Saving state to" << m_userSettings.fileName();
            m_userSettings.sync();
        });
        loadState();
    }
}

void QGreeterContext::loadState() {
    qCDebug(greeterContext) << "Loading state from" << m_userSettings.fileName();

    if (Settings::General::RememberLastUser) {
        if (auto value = m_userSettings.value("User"); value.isValid()) {
            setUsername(value.toString());
        }
    }

    if (Settings::General::RememberLastSession) {
        auto key = m_userSettings.value("Session").toString();
        auto type =
            m_userSettings.value("Type", SessionModel::Wayland).value<SessionModel::SessionType>();
        if (key.length() == 0) {
            return;
        }

        if (auto idx = m_sessionModel->findSession(type, key); idx != QModelIndex()) {
            setCurrentSession(idx.row());
        }
    }
}

SessionModel *QGreeterContext::sessionModel() const {
    return m_sessionModel;
}

const QString &QGreeterContext::hostname() const {
    return m_hostname;
}

const QString &QGreeterContext::username() const {
    return m_username;
}

void QGreeterContext::setUsername(const QString &username) {
    if (m_username != username) {
        qCDebug(greeterContext) << "set username to " << username;
        if (Settings::General::RememberLastUser) {
            m_userSettings.setValue("User", username);
        }
        emit usernameChanged(m_username = username);
    }
}

int QGreeterContext::currentSession() const {
    return m_currentSession;
}

void QGreeterContext::setCurrentSession(int session) {
    if (m_currentSession != session) {
        emit currentSessionChanged(m_currentSession = session);
    }
}

void QGreeterContext::respond(QString response) {
    /* Convert value type for QML */
    if (response.length() > 0) {
        QGreetd::QGreetdClient::respond(response);
    } else {
        QGreetd::QGreetdClient::respond();
    }
}

void QGreeterContext::startSession(QModelIndex session) {
    QStringList cmd{m_sessionModel->data(session, SessionModel::ExecRole).value<QString>()};
    QStringList env;
    QString type = "XDG_SESSION_TYPE=";
    auto sessionType =
        m_sessionModel->data(session, SessionModel::TypeRole).value<SessionModel::SessionType>();

    if (Settings::General::RememberLastSession) {
        m_userSettings.setValue("Type", sessionType);
        m_userSettings.setValue("Session", m_sessionModel->data(session, SessionModel::KeyRole));
        if (m_currentCommand != cmd[0]) {
            m_userSettings.setValue("CustomCommand", m_currentCommand);
        }
    }

    switch (sessionType) {
    case SessionModel::Wayland:
        type += "wayland";
        break;
    case SessionModel::X11: {
        type += "x11";
        /*
         * XXX: hack because startx wants absolute path to the binary.
         *
         * We wrap the command as `/usr/bin/startx /usr/bin/env cmd` to defer the the path
         * resolution until the user's ~/.profile is loaded.
         */
        auto env = QStandardPaths::findExecutable("env");
        auto startx = QStandardPaths::findExecutable("startx");
        if (!env.length() || !startx.length()) {
            qCWarning(greeterContext) << "The files required for starting X11 session are missing";
            break;
        }

        cmd.push_front(env);
        cmd.push_front(startx);
        break;
    }
    default:
        type += "tty";
        break;
    }
    env.push_back(type);
    env.push_back("XDG_SESSION_DESKTOP=" +
                  m_sessionModel->data(session, SessionModel::KeyRole).toString());

    return startSession(cmd, env);
}

void QGreeterContext::onAuthenticationComplete() {
    if (m_currentSession >= 0) {
        return startSession(m_sessionModel->index(m_currentSession, 0));

    } else if (m_currentCommand.length() > 0) {
        return startSession({m_currentCommand});

    } else {
        return startSession({"/bin/sh"});
    }
}
} // namespace QmlGreet
