/*
 * Copyright (C) 2022  Aleksei Bavshin
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
#include "qmlgreeter.hpp"
#include "config.h"
#include "qgreetercontext.hpp"
#include "util/configuration.hpp"
#include <QGuiApplication>
#include <QQmlContext>
#include <QQmlEngine>
#include <QQuickView>
#include <QScreen>

QMLGREET_DECLARE_OPTION(Wayland, LayerShell,
                        QCommandLineOption("layer-shell", "Use layer-shell for greeter surfaces."),
                        false);

QMLGREET_DECLARE_OPTION(General, Theme,
                        QCommandLineOption("theme", "Load theme from the specified <directory>.",
                                           "directory"),
                        QString());

namespace {
Q_LOGGING_CATEGORY(logcat, "qmlgreet.main");
}

namespace QmlGreet {

QmlGreeterTheme::QmlGreeterTheme(QString dirName) {
    QDir dir{dirName};
    QSettings theme{dir.absoluteFilePath("theme.conf"), QSettings::IniFormat};

    qCDebug(logcat) << "Loading theme from" << dir.absolutePath();

    QFileInfo entryFile{dir.absolutePath(), theme.value("EntryPoint", "Main.qml").toString()};
    background = theme.value("Background", "").toString();
    entryPoint = QUrl::fromLocalFile(entryFile.filePath());
    importPath = QUrl::fromLocalFile(dir.absolutePath());
    qCDebug(logcat) << "import:" << importPath << "entry:" << entryPoint
                    << "background:" << background;
}

QmlGreeter::QmlGreeter(QObject *parent) : QObject{parent} {
    if (Settings::General::Theme) {
        m_theme = QmlGreeterTheme{Settings::General::Theme.get<QString>()};
    }

    m_engine = new QQmlEngine(this);
    m_engine->addImportPath(m_theme.importPath.toString());
    m_engine->addImportPath("qrc:/");
    connect(m_engine, &QQmlEngine::quit, qGuiApp, &QGuiApplication::quit);

    /* m_context has to be destroyed after m_engine */
    m_context = new QGreeterContext{m_engine};
    connect(m_context, &QGreeterContext::done, qGuiApp, &QGuiApplication::quit);

    auto context = m_engine->rootContext();
    context->setContextProperty("background", m_theme.background);
    context->setContextProperty("greeter", m_context);
    context->setContextProperty("sessionModel", m_context->sessionModel());
}

void QmlGreeter::start() {
    if (Settings::Wayland::LayerShell) {
        qputenv("QT_WAYLAND_SHELL_INTEGRATION", "layer-shell;xdg-shell");
    }

    for (auto scrn : QGuiApplication::screens()) {
        createViewForScreen(scrn);
    }
    m_context->connectToDaemon();
}

void QmlGreeter::createViewForScreen(QScreen *screen) {
    qCDebug(logcat) << "Create view for" << screen << screen->geometry();

    auto view = new QQuickView{m_engine, nullptr};
    m_views.push_back(view);

    view->setFlags(Qt::FramelessWindowHint);
    view->setResizeMode(QQuickView::SizeRootObjectToView);
    view->setScreen(screen);
    view->setSource(m_theme.entryPoint);

    connect(qGuiApp, &QGuiApplication::screenRemoved, view, [this, screen, view](QScreen *s) {
        if (s == screen) {
            m_views.removeOne(view);
            view->deleteLater();
        }
    });

    /* xdg-shell allows passing an output hint for a fullscreen window */
    view->showFullScreen();
}

} // namespace QmlGreet
