/*
 * Copyright (C) 2022  Aleksei Bavshin
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "qtextgreeter.hpp"
#include "config.h"
#include "util/configuration.hpp"
#include "util/sessionmodel.hpp"

#include <termios.h>
#include <unistd.h>

QMLGREET_DECLARE_OPTION(QTextGreet, Command,
                        QCommandLineOption({"c", "command"},
                                           "<command> to run after authentication.", "command"),
                        "/bin/sh");
QMLGREET_DECLARE_OPTION(QTextGreet, ShowSessions,
                        QCommandLineOption("show-sessions", "Enable desktop session support."),
                        false);

namespace QmlGreet {

QString readLine(QString prompt) {
    QTextStream qout{stdout, QIODevice::WriteOnly};
    QTextStream qin{stdin, QIODevice::ReadOnly};

    qout << prompt << " ";
    qout.flush();
    return qin.readLine();
}

struct termios_disable_echo {
    struct termios saved;

    termios_disable_echo() {
        tcgetattr(STDIN_FILENO, &saved);
        struct termios value = saved;
        value.c_lflag &= ~(ECHO);
        tcsetattr(STDIN_FILENO, TCSANOW, &value);
    }

    ~termios_disable_echo() {
        tcsetattr(STDIN_FILENO, TCSANOW, &saved);
    }
};

QString readSecret(QString prompt) {
    QTextStream qout{stdout, QIODevice::WriteOnly};
    QTextStream qin{stdin, QIODevice::ReadOnly};

    qout << prompt << " ";
    qout.flush();

    struct termios_disable_echo noecho;
    return qin.readLine();
}

QModelIndex readSession(const SessionModel *model) {
    QMetaEnum sessionTypeMeta = QMetaEnum::fromType<SessionModel::SessionType>();
    QTextStream qout{stdout, QIODevice::WriteOnly};
    QTextStream qin{stdin, QIODevice::ReadOnly};

    auto rowCount = model->rowCount(QModelIndex());

    qout << "Sessions: \n";

    for (auto i = 0; i < rowCount; ++i) {
        auto idx = model->index(i, 0);
        auto type = model->data(idx, SessionModel::TypeRole);
        qout << (i + 1) << ": " << model->data(idx, Qt::DisplayRole).toString() << " ("
             << sessionTypeMeta.valueToKey(type.value<SessionModel::SessionType>()) << ")\n";
    }

    while (true) {
        int idx;
        qout << "Select a session: " << Qt::flush;
        qin >> idx;
        if (idx > 0 && idx <= rowCount) {
            return model->index(idx - 1, 0);
        }
    }
}

QTextGreeter::QTextGreeter(QObject *parent) : QObject(parent), client{} {
    using Client = QGreetd::QGreetdClient;

    connect(&client, &Client::connected, this, &QTextGreeter::onConnect);
    connect(&client, &Client::disconnected, this, &QTextGreeter::done);

    connect(&client, &Client::showPrompt, this, &QTextGreeter::onShowPrompt);
    connect(&client, &Client::showMessage, this, &QTextGreeter::onShowMessage);
    connect(&client, &Client::reset, this, &QTextGreeter::onReset);
    connect(&client, &Client::done, this, &QTextGreeter::done);
};

void QTextGreeter::start() {
    client.connectToDaemon();
}

void QTextGreeter::onConnect() {
    QString username = readLine("Username:");

    if (Settings::QTextGreet::ShowSessions) {
        auto model = client.sessionModel();
        if (model->rowCount(QModelIndex()) > 0) {
            auto session = readSession(model);
            client.setCurrentSession(session.row());
        }
    } else {
        QString cmd = Settings::QTextGreet::Command.get<QString>();
        client.setProperty("currentCommand", cmd);
        client.setCurrentSession(-1);
    }

    client.authenticate(username);
}

void QTextGreeter::onShowPrompt(QString text, PromptType type) {
    QString response;
    if (type == PromptType::Secret) {
        response = readSecret(text);
    } else {
        response = readLine(text);
    }
    client.respond(response);
}

void QTextGreeter::onShowMessage(QString text, MessageType type) {
    auto out = QTextStream{stdout, QIODevice::WriteOnly};
    if (type == MessageType::Error) {
        out << "Error: ";
    }
    out << text << "\n";
    out.flush();
}

void QTextGreeter::onReset() {
    QTimer::singleShot(500, this, SLOT(onConnect()));
}

} // namespace QmlGreet
