/*
 * Copyright (C) 2022  Aleksei Bavshin
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "configuration_p.hpp"
#include <QCommandLineParser>
#include <QLoggingCategory>
#include <QSettings>
#include <list>
#include <memory>
#include <utility>

namespace {

Q_LOGGING_CATEGORY(logcat, "qmlgreet.config");
const auto QSETTINGS_DEFAULT_GROUP_PREFIX = QLatin1String{"General/"};

} // namespace

namespace QmlGreet {

QSettingsGroupScope::QSettingsGroupScope(QSettings &settings, const QString &group)
    : m_settings{settings}, m_group{group} {
    if (m_group.startsWith(QSETTINGS_DEFAULT_GROUP_PREFIX)) {
        m_group.remove(0, QSETTINGS_DEFAULT_GROUP_PREFIX.size());
    }
    m_settings.beginGroup(group);
}

QSettingsGroupScope::~QSettingsGroupScope() {
    Q_ASSERT(m_settings.group() == m_group);
    if (m_group.length() > 0) {
        m_settings.endGroup();
    }
}

ConfigurationPrivate::ConfigurationPrivate(std::function<QSettings()> settingsFactory)
    : m_settings(settingsFactory()) {
    m_settings.setFallbacksEnabled(false);
    qCDebug(logcat) << "Reading configuration from " << m_settings.fileName();

    for (auto obj : std::as_const(getOptionList())) {
        m_cli.addOption(obj.option);
    }
}

/* QSettings was rejecting all my attempts to move it :)*/
QSettings ConfigurationPrivate::defaultSettings() {
    return QSettings{QSettings::IniFormat, QSettings::SystemScope, "qmlgreet", "qmlgreet"};
}

std::list<Option> &ConfigurationPrivate::getOptionList() {
    static std::list<Option> options;
    return options;
}

std::unique_ptr<ConfigurationPrivate> &ConfigurationPrivate::instance() {
    static std::unique_ptr<ConfigurationPrivate> m_instance =
        std::make_unique<ConfigurationPrivate>();
    return m_instance;
}

Option &Configuration::registerOption(Option &&opt) {
    return ConfigurationPrivate::getOptionList().emplace_back(opt);
}

QCommandLineParser &Configuration::createCommandLineParser() {
    return ConfigurationPrivate::instance()->m_cli;
}

QVariant Configuration::value(const QString &key, const QVariant &defaultValue) {
    return ConfigurationPrivate::instance()->m_settings.value(key, defaultValue);
}

#define RETURN(value, source)                                                                      \
    do {                                                                                           \
        qCDebug(logcat) << "Config read:" << section << key << value << "(" source ")";            \
        return value;                                                                              \
    } while (0)

template <> QVariant Option::get<QVariant>() const {
    auto &priv = *ConfigurationPrivate::instance();

    if (priv.m_cli.isSet(option)) {
        if (option.valueName() != QString()) {
            RETURN(priv.m_cli.value(option), "commandline");
        }
        RETURN(true, "commandline");
    }

    auto fullKey = section + "/" + key;
    /* Qt has a _very_ special handling for General */
    if (fullKey.startsWith(QSETTINGS_DEFAULT_GROUP_PREFIX)) {
        fullKey.remove(0, QSETTINGS_DEFAULT_GROUP_PREFIX.size());
    }
    if (auto value = priv.m_settings.value(fullKey); value.isValid()) {
        RETURN(value, "configuration");
    }
    RETURN(defaultValue, "default");
}

QStringList Option::getList(QLatin1Char separator) const {
    auto &priv = *ConfigurationPrivate::instance();

    if (priv.m_cli.isSet(option)) {
        Q_ASSERT(option.valueName() != QString());
        RETURN(priv.m_cli.values(option), "commandline");
    }

    auto fullKey = section + "/" + key;
    /* Qt has a _very_ special handling for General */
    if (fullKey.startsWith(QSETTINGS_DEFAULT_GROUP_PREFIX)) {
        fullKey.remove(0, QSETTINGS_DEFAULT_GROUP_PREFIX.size());
    }
    if (auto value = priv.m_settings.value(fullKey); value.isValid()) {
        auto list = value.toString().split(separator);
        RETURN(list, "configuration");
    }

    RETURN(defaultValue.toStringList(), "default");
}

template <> QString Option::get<QString>() const {
    return get().value<QString>();
}

template <> bool Option::get<bool>() const {
    return get().value<bool>();
}

Option::operator bool() const {
    return get<bool>();
}

#undef RETURN

} // namespace QmlGreet
