/*
 * Copyright (C) 2022  Aleksei Bavshin
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "sessionmodel.hpp"
#include <QDebug>
#include <QLocale>
#include <QLoggingCategory>
#include <QSettings>
#include <QStandardPaths>
#include <utility>

namespace QmlGreet {

Q_LOGGING_CATEGORY(sessionModel, "qmlgreet.session");

SessionModel::SessionModel(QObject *parent) : QAbstractListModel{parent} {}

QModelIndex SessionModel::findSession(SessionModel::SessionType type, const QString &key) {
    const auto idxs = match(index(0), SessionModel::KeyRole, key, -1, Qt::MatchExactly);
    for (auto idx : idxs) {
        if (data(idx, SessionModel::TypeRole) == type) {
            qCDebug(sessionModel) << "Found" << key << "at" << idx;
            return idx;
        }
    }
    qCDebug(sessionModel) << "Session" << key << "of type" << type << "not found";
    return QModelIndex{};
}

static bool ignoreSession(const QSettings &session) {
    if (auto hidden = session.value("Hidden").toString();
        hidden.compare("true", Qt::CaseInsensitive) == 0) {
        qCDebug(sessionModel) << "Ignore session" << session.fileName() << "as Hidden is set";
        return true;
    }

    if (auto nodisp = session.value("NoDisplay").toString();
        nodisp.compare("true", Qt::CaseInsensitive) == 0) {
        qCDebug(sessionModel) << "Ignore session" << session.fileName() << "as NoDisplay is set";
        return true;
    }

    if (auto tryExec = session.value("TryExec").toString(); tryExec.length() > 0) {
        if (QStandardPaths::findExecutable(tryExec) == QString()) {
            qCDebug(sessionModel) << "Ignore session" << session.fileName()
                                  << "as TryExec path is not found";
            return true;
        }
    }

    return false;
}

void SessionModel::loadSession(SessionType type, const QFileInfo &location) {
    QSettings sessionData{location.filePath(), QSettings::IniFormat};
#if QT_VERSION < QT_VERSION_CHECK(6, 0, 0)
    sessionData.setIniCodec("UTF-8");
#endif
    QString key = location.baseName();

    QStringList localeNames = {QLocale().name()};
    /* Strip territory from the locale identifier */
    if (auto idx = localeNames[0].indexOf('_'); idx != -1) {
        localeNames << localeNames[0].left(idx);
    }

    auto localized = [&](const auto key) {
        for (auto &locale : std::as_const(localeNames)) {
            const auto value = sessionData.value(QLatin1String("%1[%2]").arg(key, locale));
            if (value != QVariant()) {
                return value.toString();
            }
        }

        return sessionData.value(key).toString();
    };

    for (auto item : std::as_const(m_items)) {
        if (item.type == type && item.key == key) {
            return;
        }
    }

    /*
     * If multiple files have the same desktop file ID, the first one in the $XDG_DATA_DIRS
     * precedence order is used. Even if it is hidden and should be ignored.
     */
    for (auto item : std::as_const(m_hidden)) {
        if (item.type == type && item.key == key) {
            return;
        }
    }

    sessionData.beginGroup("Desktop Entry");

    ItemType newItem;
    newItem.type = type;
    newItem.key = key;
    newItem.name = localized("Name");
    newItem.comment = localized("Comment");
    newItem.exec = sessionData.value("Exec").toString();

    if (ignoreSession(sessionData)) {
        m_hidden << std::move(newItem);
    } else {
        m_items << std::move(newItem);
    }
}

void SessionModel::loadSessions(SessionType type, const QDir &location) {
    QStringList filters{"*.desktop"};
    const auto entryInfoList = location.entryInfoList(filters, QDir::Filter::Files);

    for (auto fileInfo : entryInfoList) {
        try {
            loadSession(type, fileInfo);

        } catch (const std::exception &err) {
            qCWarning(sessionModel) << "failed to parse session file" << fileInfo;
        }
    }
}

QHash<int, QByteArray> SessionModel::roleNames() const {
    QHash<int, QByteArray> roles;
    roles[Qt::DisplayRole] = "display";
    roles[Qt::ToolTipRole] = "toolTip";
    roles[KeyRole] = "key";
    roles[ExecRole] = "exec";
    roles[TypeRole] = "type";
    return roles;
}

int SessionModel::rowCount(const QModelIndex &parent) const {
    if (parent == QModelIndex()) {
        return m_items.length();
    }
    return 0;
}

QVariant SessionModel::data(const QModelIndex &index, int role) const {
    if (!index.isValid()) {
        return QVariant();
    }

    auto &item = m_items[index.row()];
    switch (role) {
    case Qt::DisplayRole:
        return item.name;
    case Qt::ToolTipRole:
        return item.comment;
    case KeyRole:
        return item.key;
    case ExecRole:
        return item.exec;
    case TypeRole:
        return item.type;
    default:
        qCWarning(sessionModel) << "unknown role requested" << role;
        break;
    }

    return QVariant();
}

} // namespace QmlGreet
